import scanpy as sc
import pickle as pkl

def save_pickle_file(file_path, data):
    with open(file_path, 'wb') as f:
        pkl.dump(data, f)

def read_pickle_file(file_path):
    with open(file_path, 'rb') as f:
        return pkl.load(f)

def get_gene_list(train_df):
    genes = train_df.iloc[:,1:-1]
    labels = train_df.iloc[:,-1:]
    train = sc.AnnData(genes, labels)
    celltypes_list = train.obs.celltype.unique().tolist()
    sc.pp.log1p(train)
    sc.tl.rank_genes_groups(train, groupby="celltype")
    # sc.pl.rank_genes_groups(train, groupby="celltype")
    top_50_DE_genes = {c: [] for c in celltypes_list}
    for c in top_50_DE_genes:
        detrain_df = sc.get.rank_genes_groups_df(train, group=c).head(50)
        top_50_DE_genes[c] = detrain_df['names'].tolist()
    return top_50_DE_genes

def save_gene_list(train_df, save_path):
    genes = train_df.iloc[:,1:-1]
    labels = train_df.iloc[:,-1:]
    train = sc.AnnData(genes, labels)
    celltypes_list = train.obs.celltype.unique().tolist()
    sc.pp.log1p(train)
    sc.tl.rank_genes_groups(train, groupby="celltype")
    # sc.pl.rank_genes_groups(train, groupby="celltype")
    top_50_DE_genes = {c: [] for c in celltypes_list}
    for c in top_50_DE_genes:
        detrain_df = sc.get.rank_genes_groups_df(train, group=c).head(50)
        top_50_DE_genes[c] = detrain_df['names'].tolist()
    save_pickle_file(save_path, top_50_DE_genes)


import pandas as pd
from scipy.stats import rankdata
import pickle as pkl
import numpy as np
import os

def read_pickle_file(file_path):
    with open(file_path, 'rb') as f:
        return pkl.load(f)

def rank_transform_exp(feats):
        '''converts features to exp ranks'''
        rows = []
        cols = list(feats.columns)
#       ind = list(feats.index)
        for ind in feats.index.to_list():
                c_r = []
                for col in cols:
                        c_r.append(feats.loc[ind,col])
                k = len(cols)
                rows.append([int(x)*np.exp(-1*(int(x)/k)) for x in rankdata(c_r)])
        feats = pd.DataFrame(rows)
        feats.columns = cols
#       feats.index = ind
        return feats

def rank_transform(feats):
        '''converts features to ranks'''
        rows = []
        cols = list(feats.columns)
#       ind = list(feats.index)
        for ind in feats.index.to_list():
                c_r = []
                for col in cols:
                        c_r.append(feats.loc[ind,col])
                rows.append([int(x) for x in rankdata(c_r)])
        feats = pd.DataFrame(rows)
        feats.columns = cols
#       feats.index = ind
        return feats

def rank_transform_de_exp(feats, path):
        '''converts DE gene features to ranks'''
        de_gene_dic = read_pickle_file(path)

    # get gene rank in each celltype
        rows = []
        cols = list(feats.columns)
#       ind = list(feats.index)
        for ind in feats.index.to_list():
                de_gene_value_dic = {x:{y:0 for y in de_gene_dic[x]} for x in de_gene_dic}
                de_gene_rank_dic = {x:{y:0 for y in de_gene_dic[x]} for x in de_gene_dic}
                c_r = []
#               print(de_gene_value_dic)
                for col in cols:
                        c_r.append(feats.loc[ind,col])
                        for x in de_gene_value_dic:
                                if col in de_gene_value_dic[x]:
                                        de_gene_value_dic[x][col] = feats.loc[ind,col]
                for x in de_gene_rank_dic:
                        de_gene_rank_dic[x] = {key: rank for rank, key in enumerate(sorted(de_gene_value_dic[x], key=de_gene_value_dic[x].get, reverse=True), 1)}

                for x in de_gene_rank_dic:
                        di = de_gene_rank_dic[x]
                        for i,col in enumerate(cols):
                                if col in di:
                                        denom = de_gene_rank_dic[x][col] * np.exp((-1*de_gene_rank_dic[x][col])/len(de_gene_rank_dic[x]))
                                        c_r[i]/=denom
                rows.append(c_r)
        feats = pd.DataFrame(rows)
        feats.columns = cols
#       feats.index = ind
        return feats

def rank_transform_de(feats, path):
        '''converts DE gene features to ranks'''
        de_gene_dic = read_pickle_file(path)

    # get gene rank in each celltype
        rows = []
        cols = list(feats.columns)
#       ind = list(feats.index)
        for ind in feats.index.to_list():
                de_gene_value_dic = {x:{y:0 for y in de_gene_dic[x]} for x in de_gene_dic}
                de_gene_rank_dic = {x:{y:0 for y in de_gene_dic[x]} for x in de_gene_dic}
                c_r = []
#               print(de_gene_value_dic)
                for col in cols:
                        c_r.append(feats.loc[ind,col])
                        for x in de_gene_value_dic:
                                if col in de_gene_value_dic[x]:
                                        de_gene_value_dic[x][col] = feats.loc[ind,col]
                for x in de_gene_rank_dic:
                        de_gene_rank_dic[x] = {key: rank for rank, key in enumerate(sorted(de_gene_value_dic[x], key=de_gene_value_dic[x].get, reverse=True), 1)}

                for x in de_gene_rank_dic:
                        di = de_gene_rank_dic[x]
                        for i,col in enumerate(cols):
                                if col in di:
                                        c_r[i]/=de_gene_rank_dic[x][col]
                rows.append(c_r)
        feats = pd.DataFrame(rows)
        feats.columns = cols
        feats.index = ind
        return feats



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

def fit_regression(E, D):
        clf = LinearRegression()
        if (np.sum(np.isnan(E)) > 1):
                print("Error")
                return 1, 0
        if (np.sum(np.isnan(D)) > 1):
                print("Error2")
                return 1, 0
        clf.fit(np.reshape(E, (E.shape[0], 1)), D)
        return (clf.coef_, clf.intercept_)

def calc_dropout(df):
        pg = []
        for col in df:
                pg.append(np.sum(df[col] == 0)/len(df))
        pg = np.array(pg)
        return pg

def get_prob_of_less_than(val, k):
        if np.random.random() < k:
                return 0
        else:
                return val
        
def inductive_dropout(df, f=0.5):
        p_g = calc_dropout(df)
        p_g[p_g <= 0] = 1e-6
        p_g[p_g >= 1] = 1 - 1e-6
        D_g = np.log(p_g/(1-p_g))
        R_g = np.array([df[col].mean() for col in list(df.columns)])
        E_g = np.log2(R_g + 1)
        beta, alpha = fit_regression(E_g, D_g)

        E_g_prime = f * E_g
        delta_g = beta*(E_g_prime - E_g)
        D_g_prime = D_g + delta_g

        p_g_prime = 1/ (1 + np.exp(-1*D_g_prime))
        R_g_prime = np.power(2, E_g_prime) - 1

        for i in range(len(df.columns)):
                df[df.columns[i]] = df[df.columns[i]].apply(lambda x : get_prob_of_less_than(x, p_g_prime[i]-p_g[i]))
                R_g_i = np.mean(df[df.columns[i]])
#               df[df.columns[i]] += R_g_i - R_g_prime[i]
                df[df.columns[i]] *= (R_g_prime[i]/(R_g_i + 1e-4))
        return df

def apply_dropout2(data, celltypes, f_list):
        """
        INPUT:
                data: dataframe containing cell-gene expression matrix
                celltype: pd series of cell types
                f: dropout concentration
        """
        new_df = pd.DataFrame(data.copy())
        celltypes = celltypes.tolist()
        new_celltypes = celltypes*(len(f_list) + 1)
        for f in f_list:
                new_data = inductive_dropout(data.copy(), f)
                new_df = pd.concat((new_df, new_data))
        new_df['celltype'] = new_celltypes
        return new_df

def apply_dropout(data, celltypes, n_iter, f = 0.8):
        """
        INPUT:
                data: dataframe containing cell-gene expression matrix
                n_iter: no of times to apply dropout
                f: dropout concentration
        """
        new_df = pd.DataFrame(data.copy())
        celltypes = celltypes.tolist()
        new_celltypes = celltypes*(n_iter + 1)
        for i in range(n_iter):
                data = inductive_dropout(data.copy(), f)
                new_df = pd.concat((new_df, data))
        new_df['celltype'] = new_celltypes
        return new_df


from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix, matthews_corrcoef, cohen_kappa_score
from sklearn.model_selection import train_test_split
import sys
from RankBoostClassifier.Utils import preprocess
from RankBoostClassifier.Utils import rankTransform
from RankBoostClassifier.Utils import dropout
from RankBoostClassifier.Utils import getDEgenes
import xgboost as xgb
from datetime import datetime
import argparse
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import os
import pandas as pd
import matplotlib.pyplot as plt
from imblearn.over_sampling import SMOTE
from sklearn.model_selection import GridSearchCV

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run XGBoost on the given dataset')
    parser.add_argument('--train', type=str, default=os.path.join('data', 'train.csv'), help='Path to the training data (in csv format)')
    parser.add_argument('--test', type=str, default=os.path.join('data', 'test.csv'), help='Path to the testing data (in csv format)')
    parser.add_argument('--out', type=str, default=os.path.join('log'), help='Path to the output directory')
    
    parser.add_argument('--experiment', type=str, default='pbmc', help='Experiment name')
    parser.add_argument('--f', type=float, default=0.85, help='Dropout rate to select')
    parser.add_argument('--n_iter', type=int, default=2, help='Number of dropout iterations')
    parser.add_argument('--use_exp_rank', type=bool, default=False, help='Use exponential rank transformation')
    parser.add_argument('--use_de', type=bool, default=True, help='Use DE gene selection')
    parser.add_argument('--de_gene_path', type=str, default=os.path.join('log', 'DEgenes'), help='Path to pickle dump of DE gene selection')
    parser.add_argument('--find_de_after', type=bool, default=False, help='Find DE genes after dropout')
    parser.add_argument('--class_balance', type=int, default=0, help='Use SMOTE for handling class imbalance (1 for yes, 0 for no; defaults to 0)')
    
    ## For reproducability of results
    np.random.seed(0)

    ## Parse the arguments
    args = parser.parse_args()
    experiment = args.experiment
    out = args.out
    train = args.train
    test = args.test
    f = args.f
    n_iter = args.n_iter
    use_exp_rank = args.use_exp_rank
    use_de = args.use_de
    de_gene_path = args.de_gene_path
    de_gene_path = de_gene_path + "_" + experiment +".pkl"
    find_de_after = args.find_de_after
    class_balance = args.class_balance
    
    file_out = open(f'{out}_{experiment}.txt', 'w')
    sys.stdout = file_out
    result_history = {}
    result_history[(f, n_iter)] = {'with_rank': {'val_accuracy':0, 'test_accuracy':0}}

    try:
        train_df = pd.read_csv(train, index_col = 0)
        test_df =  pd.read_csv(test, index_col = 0)
    except:
        print("Error reading the files")
        sys.exit(1)
    print("Starting script on Experiment: {}".format(experiment))
    print("Using dropout rate: {}".format(f))
    print("Using {} iterations of dropout".format(n_iter))
    new_cols = [col for col in test_df.columns if col != 'celltype'] + ['celltype']
    train_df = train_df[new_cols]
    test_df = test_df[new_cols]  
    
    print(train_df.shape, test_df.shape)

    if not find_de_after:
        getDEgenes.save_gene_list(train_df, de_gene_path)

    ## Dropout
    print("starting dropout...")
    non_zero_cols = [col for col in train_df.columns if (train_df[col] == 0).sum() != len(train_df)]
    train_df = train_df[non_zero_cols]
    if n_iter != 0:
        try:
            train_df = pd.read_csv('dropped_out_train_' + str(experiment) + '_f=' + str(f) + '_n_iter=' + str(n_iter) + '.csv', index_col = 'index')
        except:
            train_df = dropout.apply_dropout(train_df.iloc[:,:-1], train_df.celltype, n_iter, f)
            # train_df.to_csv('dropped_out_train_' + str(experiment) + '_f=' + str(f) + '_n_iter=' + str(n_iter) + '.csv')

    if find_de_after:
        getDEgenes.save_gene_list(train_df, de_gene_path)
    
    ## taking common genes
    print("Taking common genes...")
    final_columns = list(set(train_df.columns).intersection(set(test_df.columns)))
    print('Common columns', len(final_columns))
    final_columns = [i for i in final_columns if i != 'celltype'] + ['celltype']
    train_df = train_df[final_columns]
    test_df = test_df[final_columns]

    print("Starting preprocessing...")
    train_dic = preprocess.preprocess(train_df.iloc[:,:-1], train_df.iloc[:,-1:], min_cells=20,min_genes=50)
    test_dic = preprocess.preprocess(test_df.iloc[:,:-1],test_df.iloc[:,-1:], get_hvgs=False, min_cells=0, min_genes=0)

    train_adata = train_dic['data']
    test_data = test_dic['data']
    hvgs = train_dic['hvg']

    print("taking hvgs...")
    test_data = test_data[:, hvgs.index]
    train_df = train_adata.to_df()
    test_df = test_data.to_df()
    train_df['celltype'] = train_adata.obs.celltype.to_list()
    test_df['celltype'] = test_data.obs.celltype.to_list()

    ## Training and testing
    print("starting training...")
    train, val = train_test_split(train_df.sample(frac=1), test_size=0.2)
    train_feats = train.iloc[:,:-1]
    if use_exp_rank:
        train_ranks = rankTransform.rank_transform_exp(train_feats)
    else:
        train_ranks = rankTransform.rank_transform(train_feats)
    if use_de and use_exp_rank:
        train_ranks = rankTransform.rank_transform_de_exp(train_ranks, path = de_gene_path)
    if use_de:
        train_ranks = rankTransform.rank_transform_de(train_ranks, path = de_gene_path)

    val_feats = val.iloc[:,:-1]
    if use_exp_rank:
        val_ranks = rankTransform.rank_transform_exp(val_feats)
    else:
        val_ranks = rankTransform.rank_transform(val_feats)

    if use_de and use_exp_rank:
        val_ranks = rankTransform.rank_transform_de_exp(val_ranks, path = de_gene_path)
    if use_de:
        val_ranks = rankTransform.rank_transform_de(val_ranks, path = de_gene_path)
    X_train, Y_train = train_ranks.to_numpy(), train.celltype.to_list()
    X_val, Y_val = val_ranks.to_numpy(), val.celltype.to_list()

    X_combined = np.vstack((X_train, X_val))
    y_combined = Y_train + Y_val
    
#     oversample = SMOTE()
#     X_train_, Y_train_ = oversample.fit_resample(X_train, Y_train)
    
#     clf = xgb.XGBClassifier(max_depth=3, learning_rate=0.2, n_estimators = 500)
#     clf.fit(X_train_, Y_train_, sample_weight=None)
#     predicted_labels = clf.predict(X_val)

#     print("Results on Validation set")
#     print("MCC:", matthews_corrcoef(Y_val, predicted_labels))
#     print()
#     print("Kappa:", cohen_kappa_score(Y_val, predicted_labels))
#     print()
#     print("Classification Report")
#     print(classification_report(Y_val, predicted_labels))
#     a = classification_report(Y_val, predicted_labels, output_dict = True)
#     result_history[(f, n_iter)]['with_rank']['val_accuracy'] = a['accuracy']

#     print()
    print("Results on Test set")
    
    ## Train final model on the entire data (train + val)
    if class_balance == 1:
        oversample = SMOTE()
        X_combined, y_combined = oversample.fit_resample(X_combined, y_combined)
    print(X_combined[0],y_combined[0],type(X_combined),type(y_combined))
    clf = xgb.XGBClassifier(max_depth=3, learning_rate=0.2, n_estimators = 500)
    clf.fit(X_combined, y_combined, sample_weight=None)

    # parameters = {'n_estimators': [100, 200, 300, 400, 500], 'max_depth': [1, 2, 3, 4],  'grow_policy': ['depthwise', 'lossguide'], 'learning_rate': [0.1, 0.5, 1.0], 'subsample': [0.6, 0.8, 1.0]}
    # xgb_classifier = xgb.XGBClassifier(random_state = 0)
    # clf = GridSearchCV(xgb_classifier, parameters)
    # clf.fit(X_combined, y_combined)
    # print("Grid search parameters:")
    # print(clf)
    # print()
    
    test_feats = test_df.iloc[:,:-1]
    if use_exp_rank:
        test_ranks = rankTransform.rank_transform_exp(test_feats)
    else:
        test_ranks = rankTransform.rank_transform(test_feats)
    if use_de and use_exp_rank:
        test_ranks = rankTransform.rank_transform_de_exp(test_ranks, path = de_gene_path)
    if use_de:
        test_ranks = rankTransform.rank_transform_de(test_ranks, path = de_gene_path)
        
    test_ranks['celltype'] = test_df['celltype'].to_list()
    X_test, Y_test = test_ranks.iloc[:,:-1].to_numpy(), test_ranks.celltype.to_list()
    predicted_labels_test = clf.predict(X_test)
    unique_labels = list(set(list(set(Y_test)) + list(set(predicted_labels_test))))
    unique_labels.sort()
    array = confusion_matrix(Y_test, predicted_labels_test, normalize='true')
    array = array.round(2)
    print("Confusion matrix:")
    print(array)
    print()
    print("MCC:", matthews_corrcoef(Y_test, predicted_labels_test))
    print()
    print("Kappa:", cohen_kappa_score(Y_test, predicted_labels_test))
    print()
    print("Classification Report")
    print(classification_report(Y_test, predicted_labels_test))
    a = classification_report(Y_test, predicted_labels_test, output_dict = True)
    result_history[(f, n_iter)]['with_rank']['test_accuracy'] = a['accuracy']
    
    df_cm = pd.DataFrame(array, index = unique_labels, columns = unique_labels)
    plt.figure(figsize = (10,7))
    sns.heatmap(df_cm, annot=True, cmap="YlGnBu")
    plt.savefig(f'{out}_{datetime.now()}_{experiment}_confusion_matrix.png')
    


## create setup file for creating pypi package

import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '0.11'
PACKAGE_NAME = 'RankBoostClassifier'
AUTHOR = 'Suchet Aggarwal, Asmit Kumar Singh, Namrata Bhattacharya, Debarka Sengupta'
AUTHOR_EMAIL = 'suchet18105@iiitd.ac.in, asmit18025@iiitd.ac.in, namrata.bhattacharya@hdr.qut.edu.au, debarka@iiitd.ac.in'
URL = 'https://github.com/cellsemantics/boosted_rank_classifier'

LICENSE = 'Apache License 2.0'
DESCRIPTION = 'To add'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

# try:
#     with open('requirements.txt') as f:
#         required = f.read().splitlines()
# except:
#     ## if requirements.txt is not present, throw error
#     required = []
#     raise Exception("Missing requirements")

required = ["imbalanced_learn==0.10.1",
"imblearn==0.0",
"matplotlib==3.6.3",
"numpy==1.23.5",
"pandas==1.5.3",
"scanpy==1.9.1",
"scikit_learn==1.2.1",
"scipy==1.10.0",
"seaborn==0.12.2",
"xgboost==1.7.3"]

INSTALL_REQUIRES = required

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )


import scanpy as sc
  
def preprocess(genes, labels, min_genes = 200, min_cells = 20, target_sum = 1e6, n_top_genes = 3000, max_value = 10, get_hvgs=True, scale_and_hvgs = True):
        """
        INPUT:
        file_path: path to .h5ad containing scRNA-seq
        """
        ## convert to h5ad
        adata_test = sc.AnnData(genes, labels)

        ## make var names unique
        adata_test.obs_names_make_unique()
        adata_test.var_names_make_unique()

    ## filter cells with count less than 200
        sc.pp.filter_cells(adata_test, min_genes=min_genes)

        ## filter genes with count less than 20
        sc.pp.filter_genes(adata_test, min_cells=min_cells)

        ## normalise data
        sc.pp.normalize_total(adata_test, target_sum=target_sum)

        ## LogNormalise
        if not(scale_and_hvgs):
                return {'data':adata_test}

        if get_hvgs:
                ## Get HVGS
                sc.pp.log1p(adata_test)
                sc.pp.highly_variable_genes(adata_test, n_top_genes = n_top_genes)
                adata_test = adata_test[:, adata_test.var.highly_variable]

                ## scale data
                sc.pp.scale(adata_test, max_value=max_value)
                return {'data' : adata_test, 'hvg': adata_test.var.highly_variable}

        ## scale data
        sc.pp.scale(adata_test, max_value=max_value)
        return {'data':adata_test}


def get_HVGS(genes, labels, n_top_genes = 3000):
        ## convert to h5ad
        adata_test = sc.AnnData(genes, labels)
        sc.pp.highly_variable_genes(adata_test, n_top_genes = n_top_genes)
        adata_test = adata_test[:, adata_test.var.highly_variable]
        return {'data' : adata_test, 'hvg': adata_test.var.highly_variable}